# Groups management module
import paramiko
from pySSHAdmin.utilities import _check_connection


def add_group(host, user, passw, group_name, GID=None, suppress=False):
    """Add group on remote host

        Parameters:

        host (String): hostname or IP of remote server
        user (String): valid username from remote server
        passwd (String): password associated with username
        group_name (String): group name to create
        GID (integer): optional
        suppress (boolean): suppresses exit code recognition print (default False)

        Returns:

        int: exit code 0 (group created) or -1 in case of error
    """
    # Check connection first
    if _check_connection(host=host, user=user, passw=passw, suppress=True) != 0:
        if not suppress:
            print("Connection to host failed!")
        return -1

    # Declare before try
    response = ""
    exit_status = -1

    try:
        ssh = paramiko.SSHClient()
        ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        ssh.connect(hostname=host, port=22, username=user, password=passw)
        if GID != None:
            _ssh_stdin, _ssh_stdout, _ssh_stderr = ssh.exec_command(
                "sudo groupadd --gid {} {} \n".format(GID, group_name), get_pty=True)
        else:
            _ssh_stdin, _ssh_stdout, _ssh_stderr = ssh.exec_command("sudo groupadd {} \n".format(group_name),
                                                                    get_pty=True)
        exit_status = _ssh_stdout.channel.recv_exit_status()
    except paramiko.SSHException as exception:
        if not suppress: print(exception)
        return -1
    finally:
        ssh.close()

    if exit_status == 0:
        response = "Done"
    elif exit_status == 4:
        response = "GID not unique"
        exit_status = -1
    elif exit_status == 9:
        response = "Group name not unique"
        exit_status = -1

    if not suppress: print(response)

    return exit_status


def remove_group(host, user, passw, group_name, suppress=False):
    """Remove group on remote host

        Parameters:

        host (String): hostname or IP of remote server
        user (String): valid username from remote server
        passwd (String): password associated with username
        group_name (String): group name to delete
        suppress (boolean): suppresses exit code recognition print (default False)

        Returns:

        int: exit code 0 (group created) or -1 in case of error
    """
    # Check connection first
    if _check_connection(host=host, user=user, passw=passw, suppress=True) != 0:
        if not suppress:
            print("Connection to host failed!")
        return -1

    # Declare before try
    response = ""
    exit_status = -1

    try:
        ssh = paramiko.SSHClient()
        ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        ssh.connect(hostname=host, port=22, username=user, password=passw)
        _ssh_stdin, _ssh_stdout, _ssh_stderr = ssh.exec_command("sudo groupdel {} \n".format(group_name), get_pty=True)
        exit_status = _ssh_stdout.channel.recv_exit_status()
    except paramiko.SSHException as exception:
        if not suppress: print(exception)
        return -1
    finally:
        ssh.close()

    if exit_status == 0:
        response = "Done"
    elif exit_status == 6:
        response = "Group does not exist"
        exit_status = -1
    elif exit_status == 8:
        response = "Can't remove user's primary group"
        exit_status = -1

    if not suppress: print(response)

    return exit_status


def rename_group(host, user, passw, group_name, new_group_name, suppress=False):
    """Change group name on remote host

        Parameters:

        host (String): hostname or IP of remote server
        user (String): valid username from remote server
        passwd (String): password associated with username
        group_name (String): group name to change
        new_group_name (String): new group name
        suppress (boolean): suppresses exit code recognition print (default False)

        Returns:

        int: exit code 0 (group created) or -1 in case of error
    """
    # Check connection first
    if _check_connection(host=host, user=user, passw=passw, suppress=True) != 0:
        if not suppress:
            print("Connection to host failed!")
        return -1

    # Declare before try
    response = ""
    exit_status = -1

    try:
        ssh = paramiko.SSHClient()
        ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        ssh.connect(hostname=host, port=22, username=user, password=passw)
        _ssh_stdin, _ssh_stdout, _ssh_stderr = ssh.exec_command(
            "sudo groupmod -n {} {} \n".format(new_group_name, group_name), get_pty=True)
        exit_status = _ssh_stdout.channel.recv_exit_status()
    except paramiko.SSHException as exception:
        if not suppress: print(exception)
        return -1
    finally:
        ssh.close()

    if exit_status == 0:
        response = "Done"
    elif exit_status == 4 or exit_status == 6:
        response = "Specified group doesn't exist"
        exit_status - 1
    elif exit_status == 9:
        response = "Can't change - group name already in use"
        exit_status = -1

    if not suppress: print(response)

    return exit_status
