# Updates module
import paramiko
from pySSHAdmin.utilities import _check_connection


def update_YUM(host, user, passw, suppress=False):
    """Perform update of the OS by default package manager yum

        Parameters:

        host (String): hostname or IP of remote server
        user (String): valid username from remote server
        passwd (String): password associated with username
        suppress (boolean): suppresses exit code recognition print (default False)

        Returns:

        int: exit code 0 (valid) or -1 in case of error
    """
    # Check connection first
    if _check_connection(host=host, user=user, passw=passw, suppress=True) != 0:
        if not suppress:
            print("Connection to host failed!")
        return -1

    # Declare before try
    response = ""
    exit_status = -1

    try:
        ssh = paramiko.SSHClient()
        ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        ssh.connect(hostname=host, port=22, username=user, password=passw)
        _ssh_stdin, _ssh_stdout, _ssh_stderr = ssh.exec_command("sudo yum update -y \n", get_pty=True)
        exit_status = _ssh_stdout.channel.recv_exit_status()

        if exit_status == 0:
            response = "Done!"
        elif exit_status == 127:
            response = "Command yum (default package manager for YUM/Centos was not found)!"
            exit_status = -1
        else:
            if not suppress: print("Error {}".format(exit_status))
            response = _ssh_stderr.read()
            exit_status = -1
    except paramiko.SSHException as exception:
        if not suppress: print(exception)
        return -1
    finally:
        ssh.close()

    if not suppress:
        print(response)

    return exit_status


def update_APTGET(host, user, passw, suppress=False):
    """Perform update of the OS by default package manager apt-get
    
        Parameters:

        host (String): hostname or IP of remote server
        user (String): valid username from remote server
        passwd (String): password associated with username
        suppress (boolean): suppresses exit code recognition print (default False

        Returns:

        int: exit code 0 (valid) or -1 in case of error
    """
    # Check connection first
    if _check_connection(host=host, user=user, passw=passw, suppress=True) != 0:
        if not suppress:
            print("Connection to host failed!")
        return -1

    # Declare before try
    response = ""
    exit_status = -1

    try:
        ssh = paramiko.SSHClient()
        ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        ssh.connect(hostname=host, port=22, username=user, password=passw)
        _ssh_stdin, _ssh_stdout, _ssh_stderr = ssh.exec_command(
            "sudo apt-get -y update && sudo apt-get -y upgrade && sudo apt-get -y dist-upgrade\n", get_pty=True)
        exit_status = _ssh_stdout.channel.recv_exit_status()
        response = ""
        if exit_status == 0:
            response = "Done!"
        elif exit_status == 127:
            response = "Command apt-get (default package manager for APTGET/Ubuntu was not found)!"
            exit_status = -1
        else:
            if not suppress: print("Error {}".format(exit_status))
            response = _ssh_stderr.read()
            exit_status = -1
    except paramiko.SSHException as exception:
        if not suppress: print(exception)
        return -1
    finally:
        ssh.close()

    if not suppress:
        print(response)

    return exit_status


def check_updates_YUM(host, user, passw, suppress=False):
    """Perform check for updates by default package manager yum

        Parameters:

        host (String): hostname or IP of remote server
        user (String): valid username from remote server
        passwd (String): password associated with username
        suppress (boolean): suppresses exit code recognition print (default False)

        Returns:

        int: exit code 0 (no updates), 1 (updates available) or -1 in case of error
    """
    # Check connection first
    if _check_connection(host=host, user=user, passw=passw, suppress=True) != 0:
        if not suppress:
            print("Connection to host failed!")
        return -1

    # Declare before try
    response = ""
    exit_status = -1

    try:
        ssh = paramiko.SSHClient()
        ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        ssh.connect(hostname=host, port=22, username=user, password=passw)
        _ssh_stdin, _ssh_stdout, _ssh_stderr = ssh.exec_command("sudo yum check-update \n", get_pty=True)
        exit_status = _ssh_stdout.channel.recv_exit_status()
        response = ""
        if exit_status == 0:
            # No updates available case
            response = "No packages marked for update"
        elif exit_status == 100:
            # Updates available case
            response = _ssh_stdout.readlines()
            # TODO Clean list from repo-info junk

            exit_status = 1

        elif exit_status == 127:
            # Command not found case
            response = "Command yum (default package manager for YUM/Centos was not found)!"
            exit_status = -1

        elif exit_status == 1:
            # Yum error case
            response = _ssh_stderr.readlines()
            exit_status = -1

        else:
            if not suppress: print("Error {}".format(exit_status))
            response = _ssh_stderr.read()
            exit_status = -1
    except paramiko.SSHException as exception:
        if not suppress: print(exception)
        return -1
    finally:
        ssh.close()

    if not suppress:
        if (exit_status == 1):
            # List updates in preety form
            print(*response)
        else:
            print(response)

    return exit_status


def check_updates_APTGET(host, user, passw, suppress=False):
    """Perform check for updates by default package manager apt/apt-get

        Parameters:

        host (String): hostname or IP of remote server
        user (String): valid username from remote server
        passwd (String): password associated with username
        suppress (boolean): suppresses exit code recognition print (default False)

        Returns:

        int: exit code 0 (no updates), 1 (updates available) or -1 in case of error
    """
    # Check connection first
    if _check_connection(host=host, user=user, passw=passw, suppress=True) != 0:
        if not suppress:
            print("Connection to host failed!")
        return -1

    # Declare before try
    response = ""
    exit_status = -1

    try:
        ssh = paramiko.SSHClient()
        ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        ssh.connect(hostname=host, port=22, username=user, password=passw)
        _ssh_stdin, _ssh_stdout, _ssh_stderr = ssh.exec_command("sudo apt list --upgradable \n", get_pty=True)
        exit_status = _ssh_stdout.channel.recv_exit_status()
        response = None
        if exit_status == 0:
            response = _ssh_stdout.readlines()
            response.pop(0)
            # If response is empty, apt has no updates to install
            if not response:
                exit_status = 0
                response = "No packages marked for update!"
            else:
                exit_status = 1
        else:
            if not suppress: print("Error {}".format(exit_status))
            response = _ssh_stderr.read()
            exit_status = -1
    except paramiko.SSHException as exception:
        if not suppress: print(exception)
        return -1
    finally:
        ssh.close()

    if not suppress:
        if (exit_status == 1):
            # List updates in preety form
            print(*response)
        else:
            print(response)

    return exit_status


# TODO Add methods that handle key authorization

def update_YUM_key(host, user=None, key_passw=None, keyfile=None, suppress=False):
    # TODO
    pass


def update_APTGET_key(host, user=None, key_passw=None, keyfile=None, suppress=False):
    # TODO
    pass


def check_updates_YUM_key(host, user=None, key_passw=None, keyfile=None, suppress=False):
    # TODO
    pass


def check_updates_APTGET_key(host, user=None, key_passw=None, keyfile=None, suppress=False):
    # TODO
    pass
