# Users management module
import configparser, paramiko
from pathlib import Path
from pySSHAdmin.utilities import _check_connection


def _add_sudo_no_pass(host, user, passw, user_to_modify, suppress=False):
    """Add user to sudoers file

        Parameters:

        host (String): hostname or IP of remote server
        user (String): valid username from remote server
        passwd (String): password associated with username
        user_to_modify (String): user to add to sudoers (if no passed as argument added from config file)
        suppress (boolean): suppresses exit code recognition print (default False)

        Returns:

        int: exit code 0 (success) or -1 in case of error
    """
    # Check if addition is needed
    if _check_sudo_no_pass(host=host, user=user, passw=passw, user_to_find=user_to_modify, suppress=True) == 1:
        if not suppress: print("User entry found in sudoers file, nothing to do")
        return -1

    command = """
                sudo sh -c "echo 'username ALL=(ALL) NOPASSWD: ALL' >> /etc/sudoers" \n
            """
    command = command.replace("username", user_to_modify)

    # Declare exit status before try
    exit_status = -1
    response = ""

    try:
        ssh = paramiko.SSHClient()
        ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        ssh.connect(hostname=host, port=22, username=user, password=passw)
        _ssh_stdin, _ssh_stdout, _ssh_stderr = ssh.exec_command("id -u {}".format(user_to_modify))
        exit_status = _ssh_stdout.channel.recv_exit_status()
        if exit_status == 1:
            # User was not found
            if not suppress: print("User was not found!")
            return -1
        else:
            if not suppress: print("Trying to add line in sudoers file")
            # TODO Check duplicates
            _ssh_stdin, _ssh_stdout, _ssh_stderr = ssh.exec_command(command)
            exit_status = _ssh_stdout.channel.recv_exit_status()
    except paramiko.SSHException as exception:
        if not suppress: print(exception)
        return -1
    finally:
        ssh.close()

    if exit_status != 0:
        response = "Error {}".format(exit_status)
    else:
        response = "Done adding to sudoers"

    if not suppress: print(response)
    return exit_status


def _remove_sudo_no_pass(host, user, passw, user_to_modify, suppress=False):
    """Remove user from sudoers file

        Parameters:

        host (String): hostname or IP of remote server
        user (String): valid username from remote server
        passwd (String): password associated with username
        user_to_modify (String): user to remove from sudoers (if no passed as argument added from config file)
        suppress (boolean): suppresses exit code recognition print (default False)

        Returns:

        int: exit code 0 (success) or -1 in case of error
    """

    # Check if removal is needed
    if _check_sudo_no_pass(host=host, user=user, passw=passw, user_to_find=user_to_modify, suppress=True) == 0:
        if not suppress: print("User entry not found in sudoers file, nothing to do")
        return -1

    # Declare exit status before try
    exit_status = -1
    response = ""

    try:
        ssh = paramiko.SSHClient()
        ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        ssh.connect(hostname=host, port=22, username=user, password=passw)
        if not suppress: print("Trying to remove line from sudoers file")
        # Check if not duplicating entry
        command = """
            sudo sed -i 's/username ALL=(ALL) NOPASSWD: ALL//' /etc/sudoers
        """
        command = command.replace("username", user_to_modify)
        _ssh_stdin, _ssh_stdout, _ssh_stderr = ssh.exec_command(command)
        exit_status = _ssh_stdout.channel.recv_exit_status()
    except paramiko.SSHException as exception:
        if not suppress: print(exception)
        return -1
    finally:
        ssh.close()

    if exit_status != 0:
        response = "Error {}".format(exit_status)
    else:
        response = "Done deleting from sudoers"

    if not suppress: print(response)
    return exit_status


def _check_sudo_no_pass(host, user, passw, user_to_find, suppress=False):
    """Checks user existence in sudoers file

        Parameters:

        host (String): hostname or IP of remote server
        user (String): valid username from remote server
        passwd (String): password associated with username
        user_to_find (String): user to find
        suppress (boolean): suppresses exit code recognition print (default False)

        Returns:

        int: exit code 0 if user is not present in sudoers, 1 if it's present or -1 in case of error
    """

    # Declare exit status before try
    exit_status = -1
    response = ""

    try:
        ssh = paramiko.SSHClient()
        ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        ssh.connect(hostname=host, port=22, username=user, password=passw)
        _ssh_stdin, _ssh_stdout, _ssh_stderr = ssh.exec_command(
            "sudo cat /etc/sudoers | grep -c {}".format(user_to_find))
        exit_status = _ssh_stdout.read()

    except paramiko.SSHException as exception:
        if not suppress: print(exception)
        return -1
    finally:
        ssh.close()

    if exit_status != b'0\n':
        response = "User entry in sudoers found!"
        exit_status = 1
    else:
        response = "User entry not found in sudoers"
        exit_status = 0

    if not suppress: print(response)
    return exit_status


def create_user_from_config(host, user, passw, UID=None, sudo_no_pass=False, suppress=False):
    """Perform creating user (with details in config file) on remote

        Parameters:

        host (String): hostname or IP of remote server
        user (String): valid username from remote server
        passwd (String): password associated with username
        UID (int): specified user id
        sudo_no_pass (boolean): grant user executing sudo commands without password
        suppress (boolean): suppresses exit code recognition print (default False)

        Returns:

        int: exit code 0 (success) or -1 in case of error
    """
    # Check connection first
    if _check_connection(host=host, user=user, passw=passw, suppress=True) != 0:
        if not suppress: print("Connection to host failed!")
        return -1

    # Read config for user and password
    config_file_path = Path(__file__).parent.joinpath("config.cfg").resolve()
    config = configparser.ConfigParser()
    config.read(config_file_path)

    # Declare exit status before try
    exit_status = -1
    exit_status_passwd = -1
    response = ""

    try:
        ssh = paramiko.SSHClient()
        ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        ssh.connect(hostname=host, port=22, username=user, password=passw)
        if (UID == None):
            if not suppress: print("Trying to create user")
            _ssh_stdin, _ssh_stdout, _ssh_stderr = ssh.exec_command(
                "sudo useradd -m {}".format(config.get('AUTH', 'USER')))
        else:
            if not suppress: print("Trying to create user with specified UID")
            _ssh_stdin, _ssh_stdout, _ssh_stderr = ssh.exec_command(
                "sudo useradd -m -u {} {}".format(UID, config.get('AUTH', 'USER')))
        exit_status = _ssh_stdout.channel.recv_exit_status()
        # If first command was done correctly
        if exit_status == 0:
            response = "User was created "
            # Create password for created user
            if not suppress: print("Setting password for created user")
            _ssh_stdin, _ssh_stdout, _ssh_stderr = ssh.exec_command("sudo passwd {}".format(config.get('AUTH', 'USER')))
            _ssh_stdin.write("{}\n".format(config.get('AUTH', 'PASSWORD')))
            _ssh_stdin.flush()
            _ssh_stdin.write("{}\n".format(config.get('AUTH', 'PASSWORD')))
            _ssh_stdin.flush()
            exit_status_passwd = _ssh_stdout.channel.recv_exit_status()
            if exit_status_passwd == 0:
                response += "and password was set"
            else:
                response = "User created but passwd command failed, deleting created user"
                _ssh_stdin, _ssh_stdout, _ssh_stderr = ssh.exec_command(
                    "sudo userdel {}".format(config.get('AUTH', 'USER')))

        elif exit_status == 1:
            response = "Can't update password file"
            exit_status = -1
        elif exit_status == 4:
            # No unique UID
            response = "User specified UID is not unique"
            exit_status = -1
        elif exit_status == 9:
            # Username already in use
            response = "User with specified username already exists"
            exit_status = -1
        elif exit_status == 14:
            # Problems with SELinux
            response = "Problems with SELinux user mapping"
            exit_status = -1
        else:
            # General problems
            response = "Error {}".format(exit_status)

        # If needed add to sudoers file
        if sudo_no_pass:
            sudoers = _add_sudo_no_pass(host=host, user=user, passw=passw, user_to_modify=config.get('AUTH', 'USER'),
                                        suppress=True)
            if sudoers == -1:
                if not suppress: print("Error on appending to sudoers file")
                return -1
            else:
                if not suppress: print("Appended user to sudoers file")


    except paramiko.SSHException as exception:
        if not suppress: print(exception)
        return -1
    finally:
        ssh.close()

    if not suppress: print(response)

    return exit_status


def create_user(host, user, passw, user_to_create, passw_to_create, UID=None, suppress=False):
    """Perform creating user on remote

        Parameters:

        host (String): hostname or IP of remote server
        user (String): valid username from remote server
        passwd (String): password associated with username
        user_to_create (String): username to create
        passw_to_create (String): password to creating username
        UID (int): specified user id
        suppress (boolean): suppresses exit code recognition print (default False)

        Returns:

        int: exit code 0 (success) or -1 in case of error
    """
    # Check connection first
    if _check_connection(host=host, user=user, passw=passw, suppress=True) != 0:
        if not suppress: print("Connection to host failed!")
        return -1

    # Declare exit status before try
    exit_status = -1
    exit_status_passwd = -1
    response = ""

    try:
        ssh = paramiko.SSHClient()
        ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        ssh.connect(hostname=host, port=22, username=user, password=passw)
        if (UID == None):
            if not suppress: print("Trying to create user")
            _ssh_stdin, _ssh_stdout, _ssh_stderr = ssh.exec_command("sudo useradd -m {}".format(user_to_create))
        else:
            if not suppress: print("Trying to create user with specified UID")
            _ssh_stdin, _ssh_stdout, _ssh_stderr = ssh.exec_command(
                "sudo useradd -m -u {} {}".format(UID, user_to_create))

        exit_status = _ssh_stdout.channel.recv_exit_status()

        # If first command was done correctly
        if exit_status == 0:
            response = "User was created "
            # Create password for created user
            if not suppress: print("Setting password for created user")
            _ssh_stdin, _ssh_stdout, _ssh_stderr = ssh.exec_command("sudo passwd {}".format(user_to_create))
            _ssh_stdin.write("{}\n".format(passw_to_create))
            _ssh_stdin.flush()
            _ssh_stdin.write("{}\n".format(passw_to_create))
            _ssh_stdin.flush()
            exit_status_passwd = _ssh_stdout.channel.recv_exit_status()
            if exit_status_passwd == 0:
                response += "and password was set"
            else:
                response = "User created but passwd command failed, deleting created user"
                _ssh_stdin, _ssh_stdout, _ssh_stderr = ssh.exec_command("sudo userdel {}".format(user_to_create))

        elif exit_status == 1:
            response = "Can't update password file"
            exit_status = -1
        elif exit_status == 4:
            # No unique UID
            response = "User specified UID is not unique"
            exit_status = -1
        elif exit_status == 9:
            # Username already in use
            response = "User with specified username already exists"
            exit_status = -1
        elif exit_status == 14:
            # Problems with SELinux
            response = "Problems with SELinux user mapping"
            exit_status = -1
        else:
            # General problems
            response = "Error {}".format(exit_status)

    except paramiko.SSHException as exception:
        if not suppress: print(exception)
        return -1
    finally:
        ssh.close()

    if not suppress: print(response)

    return exit_status


def delete_user_from_config(host, user, passw, suppress=False):
    """Perform deleting user (with username from config file) on remote

        Parameters:

        host (String): hostname or IP of remote server
        user (String): valid username from remote server
        passwd (String): password associated with username
        suppress (boolean): suppresses exit code recognition print (default False)

        Returns:

        int: exit code 0 (success) or -1 in case of error
    """
    # Check connection first
    if _check_connection(host=host, user=user, passw=passw, suppress=True) != 0:
        if not suppress: print("Connection to host failed!")
        return -1

    # Read config for user and password
    config_file_path = Path(__file__).parent.joinpath("config.cfg")
    config = configparser.ConfigParser()
    config.read(config_file_path)

    # Declare exit status before try
    exit_status = -1
    response = ""

    try:
        ssh = paramiko.SSHClient()
        ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        ssh.connect(hostname=host, port=22, username=user, password=passw)
        _ssh_stdin, _ssh_stdout, _ssh_stderr = ssh.exec_command("id -u {}".format(config.get('AUTH', 'USER')))
        exit_status = _ssh_stdout.channel.recv_exit_status()
        if exit_status == 1:
            # User was not found
            if not suppress: print("User was not found!")
            return -1
        _ssh_stdin, _ssh_stdout, _ssh_stderr = ssh.exec_command("sudo userdel {}".format(config.get('AUTH', 'USER')))
        exit_status = _ssh_stdout.channel.recv_exit_status()

        # Check if user was added to sudoers and delete when needed
        rem = _remove_sudo_no_pass(host=host, user=user, passw=passw, user_to_modify=config.get('AUTH', 'USER'),
                                   suppress=True)
        if rem == 0:
            if not suppress: print("Removed from sudoers")

    except paramiko.SSHException as exception:
        if not suppress: print(exception)
        return -1
    finally:
        ssh.close()

    if exit_status == 0:
        response = "User deleted"
    elif exit_status == 1:
        response = "Can't update password file"
        exit_status = -1
    elif exit_status == 6:
        # Not found
        response = "User with this username does not exist on remote"
        exit_status = -1
    elif exit_status == 8:
        # Currently logged in
        response = "User with specified username is currently logged in, cannot delete"
        exit_status = -1
    elif exit_status == 12:
        # Problems with removing home dir
        response = "Problems with deleting home dir, aborting"
        exit_status = -1
    else:
        # General problems
        response = "Error - {}".format(exit_status)

    if not suppress: print(response)

    return exit_status


def delete_user(host, user, passw, user_to_delete, r_flag=False, suppress=False):
    """Perform deleting user on remote

        Parameters:

        host (String): hostname or IP of remote server
        user (String): valid username from remote server
        passwd (String): password associated with username
        user_to_delete (String): name of user to delete
        r_flag (boolean): works as -p flag saves user entry in /etc/passwd
        suppress (boolean): suppresses exit code recognition print (default False)

        Returns:

        int: exit code 0 (success) or -1 in case of error
    """
    # Check connection first
    if _check_connection(host=host, user=user, passw=passw, suppress=True) != 0:
        if not suppress: print("Connection to host failed!")
        return -1

    # Declare exit status before try
    exit_status = -1
    response = ""

    try:
        ssh = paramiko.SSHClient()
        ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        ssh.connect(hostname=host, port=22, username=user, password=passw)
        if r_flag:
            _ssh_stdin, _ssh_stdout, _ssh_stderr = ssh.exec_command("sudo userdel -r {}".format(user_to_delete))
        else:
            _ssh_stdin, _ssh_stdout, _ssh_stderr = ssh.exec_command("sudo userdel {}".format(user_to_delete))
        exit_status = _ssh_stdout.channel.recv_exit_status()
        response = "User {} deleted".format(user_to_delete)
    except paramiko.SSHException as exception:
        if not suppress: print(exception)
        return -1
    finally:
        ssh.close()

    if exit_status == 1:
        response = "Cannot update password file"
    elif exit_status == 2:
        response = "Invalid syntax"
        exit_status = -1
    elif exit_status == 6:
        response = "User was not found"
        exit_status = -1
    elif exit_status == 8:
        response = "User currently logged in"
        exit_status = -1
    elif exit_status == 12:
        response = "Cannot delete user's home dir"
        exit_status = -1

    if not suppress:
        print(response)

    return exit_status


def add_to_group(host, user, passw, user_to_add, group, suppress=False):
    """Perform adding user to group on remote

        Parameters:

        host (String): hostname or IP of remote server
        user (String): valid username from remote server
        passwd (String): password associated with username
        user_to_add (String): username to add to group
        group (String/list): group or groups names
        suppress (boolean): suppresses exit code recognition print (default False)

        Returns:

        int: exit code 0 (success) or -1 in case of error
    """

    # Check connection first
    if _check_connection(host=host, user=user, passw=passw, suppress=True) != 0:
        if not suppress: print("Connection to host failed!")
        return -1

    # Declare exit status and response before try
    exit_status = -1
    exit_status_step = []
    response = ""

    try:
        ssh = paramiko.SSHClient()
        ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        ssh.connect(hostname=host, port=22, username=user, password=passw)
        # Check if group is single string or a list
        if type(group) is str:
            # Single group
            if not suppress: print("Trying add user {} to group {}".format(user_to_add, group))
            _ssh_stdin, _ssh_stdout, _ssh_stderr = ssh.exec_command(
                "sudo usermod -a -G {} {}".format(group, user_to_add))
            exit_status = _ssh_stdout.read()
            if exit_status is None:
                exit_status = 0
            else:
                exit_status = -1
        else:
            # List of groups
            for entry in group:
                if not suppress: print("Trying add user {} to group {}".format(user_to_add, entry))
                _ssh_stdin, _ssh_stdout, _ssh_stderr = ssh.exec_command(
                    "sudo usermod -a -G {} {}".format(entry, user_to_add))
                exit_status = _ssh_stdout.read()
                if exit_status is None:
                    exit_status_step.append(0)
                else:
                    exit_status_step.append(-1)

            if exit_status_step.count(-1) == 0:
                response = "Done"
                exit_status = 0
            elif exit_status_step.count(0) == 0:
                response = "Error, user was not added to any group"
                exit_status = -1
            else:
                response = "Partially correctly executed"
                exit_status = 127

    except paramiko.SSHException as exception:
        if not suppress: print(exception)
        return -1
    finally:
        ssh.close()

    if not suppress: print(response)

    return exit_status


def remove_from_group(host, user, passw, user_to_remove, group, suppress=False):
    """Perform removing user from group on remote

        Parameters:

        host (String): hostname or IP of remote server
        user (String): valid username from remote server
        passwd (String): password associated with username
        user_to_remove (String): username to remove from group
        group (String/list): group or groups names
        suppress (boolean): suppresses exit code recognition print (default False)

        Returns:

        int: exit code 0 (success) or -1 in case of error
    """
    # Check connection first
    if _check_connection(host=host, user=user, passw=passw, suppress=True) != 0:
        if not suppress: print("Connection to host failed!")
        return -1

    # Declare exit status and response before try
    exit_status = -1
    exit_status_step = []
    response = ""

    try:
        ssh = paramiko.SSHClient()
        ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        ssh.connect(hostname=host, port=22, username=user, password=passw)
        # Check if group is single string or a list
        if type(group) is str:
            # Single group
            if not suppress: print("Trying remove user {} from group {}".format(user_to_remove, group))
            _ssh_stdin, _ssh_stdout, _ssh_stderr = ssh.exec_command("sudo deluser {} {}".format(user_to_remove, group))
            exit_status = _ssh_stdout.channel.recv_exit_status()
            if exit_status is None:
                exit_status = 0
                response = "Done"
            else:
                exit_status = -1
                response = _ssh_stdout.read()
        else:
            # List of groups
            for entry in group:
                if not suppress: print("Trying remove user {} from group {}".format(user_to_remove, entry))
                _ssh_stdin, _ssh_stdout, _ssh_stderr = ssh.exec_command(
                    "sudo deluser {} {}".format(user_to_remove, entry))
                exit_status = _ssh_stdout.channel.recv_exit_status()
                if exit_status == 0:
                    exit_status_step.append(0)
                else:
                    exit_status_step.append(-1)

            if exit_status_step.count(-1) == 0:
                response = "Done"
                exit_status = 0
            elif exit_status_step.count(0) == 0:
                response = "Error, user was not removed from any group"
                exit_status = -1
            else:
                response = "Partially correctly executed"
                exit_status = 127

    except paramiko.SSHException as exception:
        if not suppress: print(exception)
        return -1
    finally:
        ssh.close()

    if not suppress: print(response)

    return exit_status


def check_if_in_group(host, user, passw, user_to_check, group, suppress=False):
    """Checks user existence in group

        Parameters:

        host (String): hostname or IP of remote server
        user (String): valid username from remote server
        passwd (String): password associated with username
        user_to_check (String): user to check existence in group
        suppress (boolean): suppresses exit code recognition print (default False)

        Returns:

        int: exit code 0 if user is not present in group, 0 if it's present or -1 in case of error
    """

    # Declare exit status before try
    exit_status = -1
    response = ""

    try:
        ssh = paramiko.SSHClient()
        ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        ssh.connect(hostname=host, port=22, username=user, password=passw)
        _ssh_stdin, _ssh_stdout, _ssh_stderr = ssh.exec_command(
            "sudo groups {} | grep -c {}".format(user_to_check, group))
        exit_status = _ssh_stdout.read()

    except paramiko.SSHException as exception:
        if not suppress: print(exception)
        return -1
    finally:
        ssh.close()

    if exit_status != b'0\n':
        response = "User entry in sudoers found!"
        exit_status = 1
    else:
        response = "User entry not found in sudoers"
        exit_status = 0

    if not suppress: print(response)
    return exit_status


delete_user_from_config(host="ttr-test-konkur.tuatara.pl", user="root", passw="passw0rd")
