# Utilities module

from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.hazmat.backends import default_backend
from pathlib import Path
import os
import paramiko
import socket
import configparser
from pySSHAdmin.users import create_user_from_config, delete_user_from_config


def generate_RSA_keys(overwrite=False, suppress=False):
    """Generate RSA key pair for this tool
    
        Parameters:

        overwrite (boolean): overwrite key files if exists (default False)
        suppress (boolean): suppresses exit code recognition print (default False)

        Returns:

        int: exit code 0 (created) or -1 in case of error
    """
    if not suppress: print("Generating key pair")
    # Exponent (signing): 0x10001 = 2^16 + 1 = 65537
    key = rsa.generate_private_key(backend=default_backend(), public_exponent=65537, key_size=2048)
    # Private key with PKCS standard syntax
    private = key.private_bytes(serialization.Encoding.PEM, serialization.PrivateFormat.PKCS8,
                                serialization.NoEncryption())
    # Public key (OpenSSH version)
    public = key.public_key().public_bytes(serialization.Encoding.OpenSSH, serialization.PublicFormat.OpenSSH)

    # Get files paths
    pub_file_path = Path(__file__).parent.joinpath("utils").joinpath("id_rsa.pub")
    prv_file_path = Path(__file__).parent.joinpath("utils").joinpath("id_rsa")

    # File existence flags
    prv_file_exists = False
    pub_file_exists = False

    ###
    # Save private key file
    ###
    try:
        # Check if file exists
        if os.path.isfile(prv_file_path):
            prv_file_exists = True

        if overwrite and prv_file_exists:
            # Overwrite
            if not suppress: print("Purging private key file!")
            prv_file = open(prv_file_path, 'wb')
            prv_file.truncate(0)
            prv_file.write(private)
            prv_file.close()
        elif not overwrite and prv_file_exists:
            # File exists but overwrite flag is not set
            if not suppress: print("Private RSA key file exists, overwriting flag not set, exiting.")
            return -1
        else:
            # Normal file creation
            with open(prv_file_path, 'wb') as f:
                f.write(private)
            f.close()
            if not suppress: print("Private key file created and saved")
    except OSError as exception:
        if not suppress:
            print("OSError occured!")
            print(exception)
        return -1

    ###
    # Save public key file
    ###
    try:
        # Check if file exists
        if os.path.isfile(pub_file_path):
            pub_file_exists = True

        if overwrite and pub_file_exists:
            # Overwrite
            if not suppress: print("Purging public key file!")
            pub_file = open(pub_file_path, 'wb')
            pub_file.truncate(0)
            pub_file.write(public)
            pub_file.close()
        elif not overwrite and pub_file_exists:
            # File exists but overwrite flag is not set
            if not suppress: print("Pubilc RSA key file exists, overwriting flag not set, exiting.")
            return -1
        else:
            # Normal file creation
            with open(pub_file_path, 'wb') as f:
                f.write(public)
            f.close()
            if not suppress: print("Public key file created and saved")
    except OSError as exception:
        if not suppress:
            print("OSError occured!")
            print(exception)
        return -1

    return 0


def exchange_key_with_server(host, user=None, passw=False, keyfile=None, suppress=False):
    """Adds key to the remote host authorized keys (also replaces key on remote host)

        Parameters:
        
        host (String): hostname or IP of remote server
        user (String): user name on remote server (uses one from config.cfg when not provided)
        passw (String): password to remote host user's account (uses one from config.cfg when not provided)
        keyfile (String): path to id_rsa.pub file (default the one generated and stored in package)
        suppress (boolean): suppresses exit code recognition print (default False)

        Returns:

        int: exit code 0 (succesful exchange) or -1 in case of error
    """
    config_file_path = Path(__file__).parent.joinpath("config.cfg")
    config = configparser.ConfigParser()
    config.read(config_file_path)

    # Check connection first
    if _check_connection(host=host, user=user, passw=passw, suppress=True) != 0:
        if not suppress:
            print("Connection to host failed!")
        return -1

    # Read user from config file
    if not user:
        user = config.get('AUTH', 'USER')

    # Read password from config file
    if not passw:
        passw = config.get('AUTH', 'PASSWORD')

    try:
        if not keyfile:
            # Open public rsa key generated or copied to project directories
            key = open(Path(__file__).parent.joinpath("utils").joinpath("id_rsa.pub")).read()
        else:
            # Open public rsa key from given path
            key = open(Path(keyfile)).read()
    except OSError as exception:
        if not suppress: print(exception)
        return -1

        # Create .ssh dir in home of user and add keyfile to authorized
    try:
        client = paramiko.SSHClient()
        client.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        client.connect(hostname=host, username=user, password=passw)
        client.exec_command('mkdir -p ~/.ssh/')
        client.exec_command('echo "%s" > ~/.ssh/authorized_keys' % key)
        client.exec_command('chmod 600 ~/.ssh/authorized_keys')
        client.exec_command('chmod 700 ~/.ssh/')
    except paramiko.SSHException as exception:
        if not suppress: print(exception)
        return -1
    finally:
        client.close()
    if not suppress: print("Done creating entry in server's authorized keys!")
    return 0


def _check_connection(host, user, passw, suppress=False):
    """Perform connection check with host
    
        Parameters:

        host (String): hostname or IP of remote server
        user (String): valid username from remote server
        passwd (String): password associated with username
        suppress (boolean): suppresses exit code recognition print (default False)

        Returns:

        int: exit code 0 (valid) or -1 in case of error
    """
    response = ""
    exit_status = 0
    try:
        ssh = paramiko.SSHClient()
        ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        ssh.connect(hostname=host, port=22, username=user, password=passw)
        # Add to known hosts
        add_to_known_hosts(host=host, suppress=True)
        response = "Connected succesfully!"
    except paramiko.SSHException as exception:
        # Authentication, general exception case
        response = exception
        exit_status = -1
    except (ConnectionError, TimeoutError) as exception:
        # Timeout case
        response = exception
        exit_status = -1
    except socket.gaierror as exception:
        response = exception
        exit_status = -1
    finally:
        ssh.close()

    # if not suppress and response:
    if not suppress:
        print(response)

    return exit_status


# Not working properly due to Paramiko BUG (SSH-Key encryption)
def _check_connection_with_key(host, user=None, key_passw=None, keyfile=None, suppress=False):
    """Perform connection check with host (using rsa key with/without password)
    
        Parameters:

        host (String): hostname or IP of remote server
        user (String): valid username from remote server
        key (String): optional path to key file (default using generated one in project structure)
        suppress (boolean): suppresses exit code recognition print (default False)

        Returns:

        int: exit code 0 (valid) or -1 in case of error
    """
    config_file_path = Path(__file__).parent.joinpath("config.cfg")
    config = configparser.ConfigParser()
    config.read(config_file_path)

    # Read user from config file
    if not user:
        user = config.get('AUTH', 'USER')

    response = ""
    exit_status = 0
    try:
        # Try to open keyfile
        try:
            if not keyfile:
                keyfile = str(Path(__file__).parent.joinpath("utils").joinpath("id_rsa"))
            else:
                if not os.path.isfile(keyfile):
                    response = "Invalid path to private RSA keyfile!"
                    return -1
        except OSError as exception:
            if not suppress: print(exception)
            return -1

        ssh = paramiko.SSHClient()
        ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        if not key_passw:
            # Key without password
            ssh.connect(hostname=host, username=user, key_filename=keyfile)
        else:
            ssh.connect(hostname=host, username=user, password=key_passw, key_filename=keyfile)
        # Add to known hosts
        add_to_known_hosts(host=host, suppress=True)
        response = "Connected succesfully!"
    except paramiko.SSHException as exception:
        # Authentication, general exception case
        response = exception
        exit_status = -1
    except (ConnectionError, TimeoutError) as exception:
        # Timeout case
        response = exception
        exit_status = -1
    except socket.gaierror as exception:
        response = exception
        exit_status = -1
    finally:
        ssh.close()

    # if not suppress and response:
    if not suppress:
        print(response)

    return exit_status


def add_to_known_hosts(host, overwrite=False, suppress=False):
    """Perform adding to known hosts file
    
        Parameters:

        host (String): hostname or IP of remote server
        overwrite (boolean): overwrites existing entry (replacing) (default False)
        suppress (boolean): suppresses exit code recognition print (default False)

        Returns:

        int: exit code 0 (added) or -1 in case of error
    """
    known_hosts_path = Path(__file__).parent.joinpath("utils").joinpath("known_hosts")
    exit_status = 0

    t = paramiko.transport.Transport('%s:%s' % (host, 22))
    t.start_client()
    key = t.get_remote_server_key()

    ###
    # Save known_hosts file
    ###
    try:
        if os.path.isfile(known_hosts_path):
            if not suppress: print("File known_hosts found!")
        else:
            # Create known_hosts file
            f = open(known_hosts_path, 'w')
            f.close()
            if not suppress: print("Created file known_hosts and saved new entry")

        hostfile = paramiko.HostKeys(filename=known_hosts_path)
        hostfile.add(hostname=host, key=key, keytype=key.get_name())
        hostfile.save(filename=known_hosts_path)
        exit_status = 0

    except OSError as exception:
        if not suppress:
            print("OSError occured!")
            print(exception)
        exit_status = -1

    return exit_status


def prepare_remote(host, user, passw, suppress=False):
    """Perform whole preparing procedure on remote host. 
       Method is using RSA keys from package file contents and username and password from config file
       Connecting, creating user from config and adding to sudoers file, and handshaking for ssh key exchange
    
        Parameters:

        host (String): hostname or IP of remote server
        user (String): user name on remote server with root privelages
        passw (String): password to remote host user's account with root privelages
        suppress (boolean): suppresses exit code recognition print (default False)

        Returns:

        int: exit code 0 (succes) or -1 in case of error
    """
    config_file_path = Path(__file__).parent.joinpath("config.cfg")
    config = configparser.ConfigParser()
    config.read(config_file_path)

    response = ""

    # Add user on remote (also write in sudoers file)
    result = create_user_from_config(host=host, user=user, passw=passw, sudo_no_pass=True, suppress=True)
    if result != 0:
        if not suppress: print("Error during creating user on remote, try do it manually")
        return -1
    else:
        result = exchange_key_with_server(host=host, user=config.get('AUTH', 'USER'),
                                          passw=config.get('AUTH', 'PASSWORD'), suppress=True)

    if result != 0:
        if not suppress: print("Error during ssh keys exchange with remote, try do it manually")
        return -1
    else:
        response = "Done"

    if not suppress: print(response)
    return result


def cleanup_remote(host, user, passw, suppress=False):
    """Perform cleanup of remote, deleting . 
       Method is using RSA keys from package file contents and username and password from config file
       Connecting, creating user from config and adding to sudoers file, and handshaking for ssh key exchange
    
        Parameters:

        host (String): hostname or IP of remote server
        user (String): user name on remote server with root privelages
        passw (String): password to remote host user's account with root privelages
        suppress (boolean): suppresses exit code recognition print (default False)

        Returns:

    """

    config_file_path = Path(__file__).parent.joinpath("config.cfg")
    config = configparser.ConfigParser()
    config.read(config_file_path)

    response = ""

    # Remove user from remote (also delete entry in sudoers file)
    result = delete_user_from_config(host=host, user=user, passw=passw, suppress=True)
    if result != 0:
        if not suppress: print("Error removing user on remote, try do it manually")
        return -1
    else:
        response = "Done"

    if not suppress: print(response)
    return result


#### ---- Manipulate config file methods

def set_username_in_config(username):
    """Replace username in config file

        Parameters:

        username (String): desired username to change in file 

        Returns:

        int: exit code 0 (replaced), 1 when nothing to change or -1 in case of error   
    """
    config_file_path = Path(__file__).parent.joinpath("config.cfg")
    config = configparser.ConfigParser()
    config.read(config_file_path)

    if (config.get('AUTH', 'USER') == username):
        return 1

    try:
        config.set("AUTH", "USER", username)
        config.write(open(Path(__file__).parent.joinpath("config.cfg"), "w"))
        return 0
    except OSError as exception:
        print(exception)
        return -1


def get_username_from_config():
    """Get username from config file

        Returns:

        string: username from config file
    """
    config_file_path = Path(__file__).parent.joinpath("config.cfg")
    config = configparser.ConfigParser()
    config.read(config_file_path)

    return config.get('AUTH', 'USER')


def set_password_in_config(password):
    """Replace password in config file

        Parameters:

        password (String): desired password to change in file 

        Returns:

        int: exit code 0 (replaced), 1 when nothing to change or -1 in case of error   
    """
    config_file_path = Path(__file__).parent.joinpath("config.cfg")
    config = configparser.ConfigParser()
    config.read(config_file_path)

    if (config.get('AUTH', 'PASSWORD') == password):
        return 1

    try:
        config.set("AUTH", "PASSWORD", password)
        config.write(open(Path(__file__).parent.joinpath("config.cfg"), "w"))
        return 0
    except OSError as exception:
        print(exception)
        return -1


def get_password_from_config():
    """Get password from config file

        Returns:

        string: password from config file
    """
    config_file_path = Path(__file__).parent.joinpath("config.cfg")
    config = configparser.ConfigParser()
    config.read(config_file_path)

    return config.get('AUTH', 'PASSWORD')
